import 'dart:async';

import 'package:meta/meta.dart';

import 'controller.dart';
import 'update.dart';

// For pagination,

typedef PaginatedFetcher<T, PaginationState>
    = Future<PaginationResponse<T, PaginationState>> Function(
        PaginationState state);

@immutable
class PaginationResponse<T, PaginationState> {
  const PaginationResponse({
    @required this.items,
    @required this.state,
  })  : assert(items != null),
        assert(state != null);

  final List<T> items;
  final PaginationState state;
}

class PaginatedController<T, PaginationState> {
  PaginatedController({
    @required this.fetcher,
    @required this.saveToCache,
    @required this.loadFromCache,
    this.initialState,
  });

  final PaginatedFetcher<T, PaginationState> fetcher;
  final ToCacheSaver<List<T>> saveToCache;
  final FromCacheLoader<List<T>> loadFromCache;
  final PaginationState initialState;

  final _controller = UpdatesController<List<T>>();
  Future<void> _currentFetcher;
  PaginationState _state;
  PaginationState get state => _state;

  CacheUpdate<List<T>> get lastUpdate => _controller.lastUpdate;
  bool get isFetching => lastUpdate?.isFetching ?? false;
  List<T> get lastItems => lastUpdate?.data;
  dynamic get lastError => lastUpdate?.error;
  StackTrace get lastStackTrace => lastUpdate?.stackTrace;

  /// A stream of [CacheUpdate]s of this controller. A new update appears
  /// everytime something new happens that's also relevant for the listener
  /// (like, new data is available or an error occurred during fetching).
  Stream<CacheUpdate<List<T>>> get updates => _controller.updates;

  /// Fetches data from the cache and the [fetcher] simultaneously and returns
  /// the final [CacheUpdate]. If you want to receive updates about events in
  /// between (like events from the cache), you should rather listen to the
  /// [updates] stream.
  Future<CacheUpdate<List<T>>> fetch() async {
    // If multiple clients called fetch simultaneously, we only fetch once.
    _currentFetcher ??= _actuallyFetchData();
    await _currentFetcher;
    _currentFetcher = null;
    return lastUpdate;
  }

  Future<void> _actuallyFetchData() async {
    if (lastItems == null) {
      _controller.add(CacheUpdate<List<T>>.loading());
    } else {
      _controller.add(CacheUpdate<List<T>>.cached(lastItems));
    }

    // Simultaneously get data from the (probably faster) cache and the
    // original source.
    await Future.wait([
      // Get data from the cache only if we don't already have it in memory.
      if (lastItems == null)
        Future.microtask(() async {
          try {
            final data = await loadFromCache();
            // If the original source was faster than the cache, we don't need
            // to do anything. Otherwise, we push an update with the cached
            // data so that it can be already displayed.
            if (isFetching) {
              _controller.add(CacheUpdate<List<T>>.cached(data));
            }
          } on NotInCacheException {
            // We just fail silently as the original data will be returned
            // soon – hopefully.
          }
        }),
      // Get data from the original source.
      () async {
        _state = initialState;
        _actuallyFetchMoreData();
      }(),
    ]);
  }

  /// Fetches more data from the original source.
  Future<CacheUpdate<List<T>>> fetchMore() async {
    // If multiple clients called fetch simultaneously, we only fetch once.
    _currentFetcher ??= _actuallyFetchMoreData();
    await _currentFetcher;
    _currentFetcher = null;
    return lastUpdate;
  }

  Future<void> _actuallyFetchMoreData() async {
    _controller.add(CacheUpdate<List<T>>.cached(lastItems));

    // Get data from the original source.
    Future.microtask(() async {
      try {
        final response = await fetcher(_state);
        final items = lastItems + response.items;
        _state = response.state;

        _controller.add(CacheUpdate<List<T>>.success(items));
        saveToCache(items);
      } catch (error, stackTrace) {
        _controller.add(CacheUpdate<List<T>>.error(
          error,
          stackTrace,
          cachedData: lastItems,
        ));
      }
    });
  }
}
